"""Hosts tests for the plugin."""
import functools
import os
import types
from pathlib import Path
from runpy import run_module
from subprocess import PIPE, check_output, run

import pytest
from tox import __version__ as tox_version

TOX_SAMPLE = """
[tox]
skipsdist = true
"""

tox3_only = pytest.mark.skipif(not tox_version.startswith("3."), reason="requires tox3")
tox4_only = pytest.mark.skipif(not tox_version.startswith("4."), reason="requires tox4")


def preserve_cwd(function):
    """Decorator for restoring cwd."""

    @functools.wraps(function)
    def decorator(*args, **kwargs):
        cwd = os.getcwd()
        try:
            return function(*args, **kwargs)
        finally:
            os.chdir(cwd)

    return decorator


@tox3_only
def test_import_hook() -> None:
    """One placeholder test."""
    # pylint: disable=import-outside-toplevel
    from tox_extra.hooks import tox_runtest_post

    assert isinstance(tox_runtest_post, types.FunctionType)


@preserve_cwd
def test_fail_if_dirty(tmp_path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Validated that it fails when drity."""
    # We need to mock sys.argv because run_module will look at it and fail
    # as the argv received of pytest would leak into our tox module call.
    monkeypatch.setattr("sys.argv", ["tox"])

    path = Path(tmp_path)
    os.chdir(path)

    # add tox.ini and .gitignore files (untracked)
    with open(Path(tmp_path) / "tox.ini", "w") as file:
        file.write(TOX_SAMPLE)

    # check running tox w/o git repo is passing
    with pytest.raises(SystemExit) as exc:
        run_module("tox", run_name="__main__")  # , alter_sys=True)
    assert exc.type == SystemExit
    assert exc.value.code == 0

    # create temp git repository
    run("git init --initial-branch=main", shell=True, check=True)
    run("git add --all", shell=True, check=True)
    run(
        "git commit --allow-empty -m 'Empty initial commit'",
        shell=True,
        check=True,
        cwd=tmp_path,
    )
    output = check_output("git status --porcelain", shell=True, universal_newlines=True)
    assert output == ""

    with open(Path(tmp_path) / ".gitignore", "w") as file:
        file.write(".tox\n")

    result = run(
        "git status --porcelain",
        shell=True,
        universal_newlines=True,
        cwd=tmp_path,
        stdout=PIPE,
        check=False,
    )
    assert result.returncode == 0
    assert "?? .gitignore" in result.stdout

    # check plugin is installed
    result = run(
        "python -m tox --version",
        shell=True,
        universal_newlines=True,
        stdout=PIPE,
        check=False,
    )
    assert result.returncode == 0
    assert "tox-extra" in result.stdout

    # check tox is failing while dirty
    # We use runpy to call tox in order to assure that coverage happens, as
    # running a subprocess would prevent it from working.
    with pytest.raises(SystemExit) as exc:
        run_module("tox", run_name="__main__", alter_sys=True)
    assert exc.type == SystemExit
    assert exc.value.code == 1

    # add untracked files
    run("git add .", shell=True, check=True)
    run("git commit -m 'Add untracked files'", shell=True, check=True)

    # check that tox is now passing
    with pytest.raises(SystemExit) as exc:
        run_module("tox", run_name="__main__", alter_sys=True)
    assert exc.type == SystemExit
    assert exc.value.code == 0



"""Tox hook implementations."""
import os
from argparse import ArgumentParser

import git

MSG_GIT_DIRTY = (
    "Git reported dirty status. "
    "Git should never report dirty status at the end of "
    "testing, regardless if status is passed, failed or aborted."
)


def is_git_dirty(path: str) -> bool:
    """Reports if the git repository at the given path is dirty."""
    if os.path.isdir(f"{path}/.git"):
        _environ = dict(os.environ)
        try:
            repo = git.Repo(os.getcwd())
            if repo.is_dirty(untracked_files=True):
                os.system("git status")
                os.system("git diff -U0 --minimal")
                return True
        finally:
            os.environ.clear()
            os.environ.update(_environ)
    return False


try:  # tox3 support
    from tox import config, hookimpl
    from tox.reporter import error

    @hookimpl
    def tox_addoption(parser: config.Parser) -> None:
        """Add a command line option to the tox parser."""
        parser.add_argument(
            "--allow-dirty",
            action="store_true",
            default=False,
            help="If it should allow git to report dirty after executing commands.",
        )

    @hookimpl
    def tox_runtest_post(venv):
        """Hook that runs after test commands."""
        allow_dirty = getattr(venv.envconfig.config.option, "allow_dirty", False)
        if not allow_dirty and is_git_dirty(venv.envconfig.config.toxinidir):
            venv.status = "failed"
            error(MSG_GIT_DIRTY)

except ImportError as exc:  # tox4 support
    print(exc)
    from typing import List

    from tox.execute import Outcome
    from tox.plugin import impl
    from tox.tox_env.api import ToxEnv
    from tox.tox_env.errors import Fail

    @impl
    def tox_add_option(parser: ArgumentParser) -> None:
        """Add a command line option to the tox parser."""
        parser.add_argument(
            "--allow-dirty",
            action="store_true",
            default=False,
            help="If it should allow git to report dirty after executing commands.",
        )

    @impl
    # pylint: disable=unused-argument
    def tox_after_run_commands(
        tox_env: ToxEnv, exit_code: int, outcomes: List[Outcome]
    ) -> None:
        """Hook that runs after test commands."""
        allow_dirty = getattr(tox_env.options, "allow_dirty", False)
        if not allow_dirty and is_git_dirty("."):
            raise Fail(MSG_GIT_DIRTY)
